local BASE = (...):match("([^%.^/^\\]+)")
--local class = require (BASE.."/middleclass")
local core = {} --class("core")
local default_style = require (BASE.."/theme/style_default")
local default_render = require (BASE.."/theme/render_default")
local layout = require (BASE.."/layout")
local widget_funcs = {
    button = require(BASE.."/widget/basic/button"),
    label = require(BASE.."/widget/basic/label"),
    shape = require(BASE .. "/widget/basic/shape"),
    arrow = require(BASE .. "/widget/basic/arrow"),
    check = require(BASE.."/widget/basic/check"),
    radio =require(BASE.."/widget/basic/radio"),
    image = require(BASE.."/widget/basic/image"),
    panel = require(BASE.."/widget/basic/panel"),
    progress = require(BASE.."/widget/basic/progress"),
    block = require(BASE.."/widget/basic/block"),
    text = require(BASE .."/widget/basic/text"),
    canvas = require(BASE .. "/widget/basic/canvas"),
    
    
    colorpicker = require(BASE.."/widget/combo/color"),
    slider = require(BASE.."/widget/combo/slider"),
    menu = require(BASE .. "/widget/combo/menu"),
    list = require(BASE .. "/widget/combo/list"),
    combo = require(BASE .."/widget/combo/combo"),
    number = require(BASE .. "/widget/combo/number"),
    tip = require(BASE.."/widget/combo/tip"),
    
    frame = require(BASE .. "/widget/container/frame"),
    tab = require(BASE.."/widget/container/tab")
    }


function core:init()
    
    self.mouse_x,self.mouse_y = 0,0
    self.active = nil
    self.active_o = nil
    self.hovered = nil
    self.hovered_o = nil
    self.hit = nil
    self.hit_o = nil
    self.focus = nil --keyboardfocus
    self.text = ""
    self.key = nil
    self.mouse_down = nil
    self.buff = {}
    self.style = default_style
    self.render = default_render
    self.layout = layout
    self.widgets = {}
    self.hang_timer = 1
    self.hang_cd = 1
    self.double_click_timer = 0
    for name,func in pairs(widget_funcs) do
        self.widgets[name] = func
    end
    return self
end

function core:doWidget(name,opt)
    return self.widgets[name](self,opt)
end

function core:keypressed(key)
    self.key = key
end

function core:textinput(text)
    self.text = (self.text or "").. text
end

function core:wheelmoved(dy)
    self.key = dy > 0 and "wheelup" or "wheeldown"
end

function core:draw()
    self:predraw()    
    for i = #self.buff,1,-1 do 
		self.buff[i]()
	end
    --core.render.drawToRender()
	self:postdraw()
end

function core:predraw()
    love.graphics.push()
    if self.hanging and self.hovered and self.hovered.tip then
        core:doWidget("tip",self.hovered)
    end
    
end

function core:postdraw()
    if not self.mouse_down then
        self.active_o = self.active 
		self.active = nil
	else
        if not self.active then
            self:releaseFocus()
        end
    end
    
    local dk = (love.mouse.isDown(1) and 1) or (love.mouse.isDown(2) and 2)
	self:updateMouse(love.mouse.getX(), love.mouse.getY(), dk)
    
    if self.hovered and self.hovered == self.hovered_o 
        and self.mouse_dx == 0 and self.mouse_dy == 0 then
        self.hang_timer = self.hang_timer - love.timer.getDelta()
        if self.hang_timer<0 then
            self.hanging = true
        else
            self.hanging = false
        end
    else
        self.hanging = false
        self.hang_timer = self.hang_cd
        self.tip_obj = nil
    end
    
    
    
	self.hovered_o, self.hovered = self.hovered, nil
    self.hit_o,self.hit = self.hit,nil
    
	self.key, self.text = nil, nil
    self.buff = {}
    love.graphics.pop()
end

--- hover ---

function core:isHovered(id)
    return self.hovered == id
end

function core:setHovered(id)
    return self.hovered ~= id
end

function core:anyHovered()
    return self.hovered ~= nil
end

function core:wasHovered(id)
    return self.hovered_o == id
end

function core:getHovered()
    return self.hovered
end

--- active ---

function core:isActive(id)
    return self.active == id
end

function core:setActive(id)
    return self.active ~= id
end

function core:anyActive()
    return self.active ~= nil
end

function core:wasActive(id)
    return self.active_o == id
end

function core:getActive()
    return self.active
end

--- hit --
function core:setHit(id)
    self.mouse_down = nil
    self.hit_o = self.hit
    self.active_o = self.active
    self.hovered = self.hovered_o
	self.hit = id
	self.active = id
	self.hovered = id
end

function core:anyHit()
	return self.hit ~= nil
end

function core:isHit(id)
	return id == self.hit
end




--- stata name -- 

function core:getStateName(id)
	if self:isActive(id) then
		return "active"
	elseif self:isHovered(id) then
		return "hovered"
	elseif self:isHit(id) then
		return "hit"
	end
	return "normal"
end

-- mouse handling
function core:mouseInRect(x,y,w,h)
	return self.mouse_x >= x and self.mouse_y >= y and
	       self.mouse_x <= x+w and self.mouse_y <= y+h
end

function core:registerMouseHit(id, hit)
	if hit then
		self.hovered = id
		if self.active == nil and self.mouse_down then
            self.active_o = self.active
			self.active = id
            if self.focus ~= self.active then
                self:releaseFocus()
            end
		end
	end
	return self:getStateName(id)
end

local text

function core:registerHitbox(id, x,y,w,h)
    local test
    local mx,my = self.mouse_x,self.mouse_y
    if id.container then
        local cx,cy,cw,ch = unpack(id.container)
        test = mx>= x and mx<=x+w and my>=y and my<=y+h
        test = mx>=cx and mx<=cx+cw and my>=cy and my<=cy+ch and test
    else
        test = mx>= x and mx<=x+w and my>=y and my<=y+h
    end
    
	return self:registerMouseHit(id,test)
end

function core:mouseReleasedOn(id)
	if not self.mouse_down and self:anyActive() and self:isHovered(id) then
		self.hit = id
		return self:getActive()
	end
	return false
end

function core:updateMouse(x, y, button)
    self.mouse_ox,self.mouse_oy = self.mouse_x, self.mouse_y
    self.mouse_dx = x - self.mouse_x ;
    self.mouse_dy = y - self.mouse_y ;
	self.mouse_x, self.mouse_y = x,y
    self.double_click_timer = self.double_click_timer - love.timer.getDelta()
    self.double_click = false
    if self.double_click_timer>0 and not button then
        self.click_once = true
    end
    
    if self.click_once and button and self.double_click_timer>0 then
        self.double_click = true
        self.click_once = false
        --print("wow")
    end
    
    if button then
        self.double_click_timer = 0.2
        self.click_once = false
    end
    self.mouse_down = button
end

function core:getMousePosition()
	return self.mouse_x, self.mouse_y
end

-- keyboard handling
function core:getKey()
	return self.key, self.text
end

function core:grabFocus(id)
	if self:isActive(id) then
		if love.system.getOS() == "Android" or love.system.getOS() == "iOS" then
			if not id then
				love.keyboard.setTextInput( false )
			else
				love.keyboard.setTextInput( true )
			end
		end
        self.focus_o = self.focus
		self.focus = id
	end
	return self:hasFocus(id)
end

function core:hadFocus(id)
    return self.focus ~= id and self.focus_o == id
end

function core:gotFocus(id)
   return  self.focus == id and self.focus_o ~= id
end

function core:hasFocus(id)
	return self.focus == id
end

function core:getFocus()
    return self.focus
end

function core:releaseFocus()
    self.focus_o = self.focus
    self.focus = nil
    if love.system.getOS() == "Android" or love.system.getOS() == "iOS" then
        love.keyboard.setTextInput( false )
    end
end

function core:keyPressedOn(id, key)
	return self:hasFocus(id) and self.key == key
end

function core:drawToMask()
    
end

-- draw
function core:registerDraw(f, args,invert)
    if invert then
        table.insert(self.buff,1,function()
            f(unpack(args))
        end)
    else
        table.insert(self.buff,function()
            f(unpack(args))
        end)
    end
end

function core:getRect(opt)
    return opt.x or 0,opt.y or 0,opt.w or 100,opt.title_h or opt.h or 100
end

function core:applyLayout(opt,conf)
    conf = conf or opt.layout
    if not conf then return end
    if conf.disable then return end
    local layout = self.layout
    if conf == "right" then
        layout:toRight(opt)
    elseif conf == "down" then
        layout:toDown(opt)
    elseif conf == "next" then
        layout:toNext(opt)
    elseif type(conf) == "table" then
        layout:setGrid(opt,conf)
    elseif string.match(conf,"(%w+)(%d+)x(%d+)") then
        local dir,w,h = string.match(conf,"(%w+)(%d+)x(%d+)")
        self:applyLayout(opt,dir)
        self:applyLayout(opt,{layout.cc,layout.cr,w,h})
    end
   
end


function core:switchToggle(opt)
    if opt.toggle == nil then return opt.state end
    if self:wasActive(opt) then opt.toggle = not opt.toggle end
    return opt.toggle and "active" or opt.state
end

function core:updateState(opt)
    opt.state = core:registerHitbox(opt, core:getRect(opt))
    opt.state = core:switchToggle(opt)
    opt.hit = core:mouseReleasedOn(opt) and opt == core:getHovered()
    opt.drag = core:mouseReleasedOn(opt) and core:getActive() and core:getActive() ~= opt
    opt.drag_from = core:getActive()
    opt.hovered = core:isHovered(opt)
    opt.entered = core:isHovered(opt) and not core:wasHovered(opt)
    opt.left = not core:isHovered(opt) and core:wasHovered(opt)
end

--[[
数据绑定：
当input submit时，data 更新数据
当input不在editing是，data赋予数据
data应为一个表
]]
function core:dataBind(dataObj,inputObj,name,dataType,cross)
    inputObj.bound = {
        obj = dataObj,
        key = name or "text",
        type = dataType or "string",
    }
    if cross then
        self:dataBind(inputObj,dataObj,name,dataType)
    end
end

function core:updateBound(opt)
    if not opt.bound then return end
    local obj = opt.bound.obj
    local type = opt.bound.type
    local key = opt.bound.key
    local value = opt.value or opt.text
    if opt.submit then
        if type == "number" then
            obj[key] = tonumber(value)
        elseif type == "string" then
            obj[key] = tostring(value)
        elseif type == "color" then
            if key == "rgb" then
                obj[1],obj[2],obj[3] = unpack(value)
            else --alpha
                obj[4] = tonumber(value)
            end
        elseif type == "color_rgb" then
            obj[key] = string.format("r: %0.2f,g: %0.2f,b: %0.2f",unpack(value))
        elseif type == "percentage" then
            obj[key] = string.format("%0.2f%%",value*100)
        end
    elseif not opt.active then
        
        if type == "color" then
            if key == "rgb" then
                opt.value[1],opt.value[2],opt.value[3] = unpack(obj)
            else --alpha
                opt.value = obj[4]
            end
        elseif type == "percentage" then
            opt.value = tonumber(string.sub(obj[key],1,-2))/100
        else
            if opt.value then
                opt.value = tonumber(obj[key])
            elseif opt.text then
                opt.text = tostring(obj[key])
            end
        end
    end
end

return core