local BASE = ...
local alim = {}
local core = require (BASE.."/core")


function alim:init()
    self.origin = {}
    self.core = core:init()
    self.layout = self.core.layout
    self.style = self.core.style
    self:loadWidget()
    return self
end

function alim:loadWidget()
    for name,wid in pairs(self.core.widgets) do
        self[name] = function(opt) return wid(self.core,opt) end
    end
end

function alim:widget(tab)
    return self[tab.name](tab.conf)
end

function alim:patch(_love)
    local _draw = _love.draw
    self.env = _love
    self.origin.draw = _draw
    _love.draw = function()
        self.core:draw()
        _ = _draw and _draw()
    end
    
    local _keypressed = _love.keypressed
    self.origin.keypressed = _keypressed
    _love.keypressed = function(k)
        if self.core:keypressed(k) then --block by nil
            _ = _keypressed and _keypressed(k)
        end
    end
    
    local _textinput = _love.textinput
    self.origin.textinput = _textinput
    _love.textinput = function(t)
        if self.core:textinput(t) then --block by nil
            _ =  _textinput and _textinput(t)
        end
    end
    
    local _wheelmoved = _love.wheelmoved
    self.origin.wheelmoved = _wheelmoved
    _love.wheelmoved = function(dx,dy)
        if self.core:wheelmoved(dy) then
           _ = _wheelmoved and _wheelmoved(dx,dy) 
        end
    end
end

function alim:depatch()
    for name,func in pairs(self.origin) do
        self.env[name] = func
    end
end

function alim:dataBind(...)
    self.core:dataBind(...)
end

return alim

