local layout = {
    buff = {},
    buff_max_size = 32,
    x = 0,y = 0, w = 0, h = 0,
    cx = 0,cy = 0,cc = 0,cr = 0,
    cs = 0,rs = 0,col = 0,row = 0,
    dynamic = true
    }
--[[
"x","y","w","h",
    "ix","iy","iw","ih",
    "row","rs","col","cs",
    "begin","mode",
    "cc","cr","dynamic"
]]

--[[
setWindow 设置外框 无参数则以当前窗体为外框
setCell 设置单元格大小， w,h,rs,cs
cutGrid 切分单元格，col,row,rs,cs
]]

function layout:reset()
   self.x = 0
   self.y = 0
   self.w = 1
   self.h = 1
   self.ix = 0
   self.iy = 0
   self.iw = 1
   self.ih = 1
   self.row = 1
   self.rs = 0
   self.col = 1
   self.cs = 0
   self.begin = false
   self.mode = nil
   self.cc = 1
   self.cr = 1
   self.dynamic = true
end

function layout:getGridPos(col,row)
    local x = self.x + self.cs + (col-1)*(self.cs + self.iw)
    local y = self.y + self.rs + (row-1)*(self.rs + self.ih)
    return x,y
end

function layout:apply(opt)
    if self.dynamic then
        opt.x,opt.y,opt.w,opt.h = 
        layout.ix,layout.iy,
        layout.iw,layout.ih
    else
        opt.x,opt.y,opt.w,opt.h = 
        opt.x or layout.ix,opt.y or layout.iy,
        opt.w or layout.iw,opt.h or layout.ih
    end
    
end

function layout:toBegin(item)
    self.begin = true
    if self.mode == "grid" then
        self.cc = 1
        self.cr = 1
        self.ix,self.iy = self:getGridPos(self.cc,self.cr)
    elseif self.mode == "free" then
        self.prev = item
        self.ix , self.iy = self.x + self.cs , self.y + self.rs
        self.iw = item.w or self.w - self.cs*2
        self.ih = item.h or self.h - self.rs*2
    end
    self:apply(item)
end

function layout:toRight(item) -- 位置及宽高根据网格
    if not self.begin then return self:toBegin(item) end
    if self.mode == "grid" then
        self.cc = self.cc + 1
        if self.cc > self.col then
            self.cc = 1
            self.cr = self.cr + 1
        end
        self.ix,self.iy = self:getGridPos(self.cc,self.cr)
    elseif self.mode == "free" then
        self.ix = self.ix + self.iw + self.cs
        self.iw = item.w or self.iw
        self.ih = item.h or self.ih
        if self.ix + self.iw + self.cs > self.x + self.w then
            self.ix = self.x + self.cs
            self.iy = self.iy + self.ih + self.rs
        end
    end
    self:apply(item)
end

function layout:toNext(item)
    if not self.begin then return self:toBegin(item) end
    if self.mode == "grid" then
        self.cc = 1
        self.cr = self.cr + 1
        self.ix,self.iy = self:getGridPos(self.cc,self.cr)
    elseif self.mode == "free" then
        self.ix = self.x + self.cs
        self.iy = self.iy + self.ih + self.rs
        self.iw = item.w or self.iw
        self.ih = item.h or self.ih
    end
    self:apply(item)
end


function layout:toDown(item)-- 位置及宽高根据网格
    if not self.begin then return self:toBegin(item) end
    if self.mode == "grid" then
        self.cr = self.cr + 1
        if self.cr > self.row then
            self.cr = 1
            self.cc = self.cc + 1
        end
        self.ix,self.iy = self:getGridPos(self.cc,self.cr)
    elseif self.mode == "free" then
        self.iy = self.iy + self.ih + self.rs
        self.iw = item.w or self.iw
        self.ih = item.h or self.ih
        if self.iy + self.ih + self.rs > self.y + self.h then
            self.iy = self.y + self.rs
            self.ix = self.ix + self.iw + self.cs
        end
    end
    self:apply(item)
end



function layout:setWindow(x,y,w,h,cs,rs)
    self:reset()
    self.x = x or 0
    self.y = y or 0
    self.w = w or love.graphics.getWidth()
    self.h = h or love.graphics.getHeight()
    self:setSpacing(cs,rs)
    self.mode = "free"
    self.begin = false
end

function layout:setSpacing(colSpacing,rowSpacing)
    self.cs = colSpacing or 4
    self.rs = colSpacing or 4
end

function layout:cutGrid(col,row,cs,rs)
    self.col = col or 1
    self.row = row or 1
    self:setSpacing(cs,rs)
    self.iw = (self.w-self.cs) / self.col - self.cs
    self.ih = (self.h-self.rs) / self.row - self.rs
    
    self.begin = false
    self.mode = "grid"
end

function layout:setCell(w,h,cs,rs)
    self.iw = w
    self.ih = h
    self:setSpacing(cs,rs)
    self.col = math.floor((self.w - self.cs)/ (self.iw + self.cs))
    self.row = math.floor((self.h - self.rs)/ (self.ih + self.rs))
    self.w = self.col * (self.w + self.cs) + self.cs
    self.h = self.row * (self.h + self.rs) + self.rs
    self.begin = false
    self.mode = "grid"
end

function layout:setGrid(item,conf)
    item.x, item.y = self:getGridPos(conf[1],conf[2])
    item.w = (conf[3] or 1)*(self.iw+self.cs) - self.cs
    item.h = (conf[4] or 1)*(self.ih+self.rs) - self.rs
    self.cc = conf[1]+(conf[3] or 1) - 1
    self.cr = conf[2]+(conf[4] or 1) - 1
    self.begin = true
end

local attr = {
    "x","y","w","h",
    "ix","iy","iw","ih",
    "row","rs","col","cs",
    "begin","mode",
    "cc","cr","dynamic"
    }

function layout:push()
    local d = {}
    for i,name in ipairs(attr) do
        d[name] = self[name]
    end
    table.insert(self.buff,d)
    if #self.buff > self.buff_max_size then error("max buff size reached") end
end

function layout:pop()
    local d = self.buff[#self.buff]
    self.buff[#self.buff] = nil
    for i,name in ipairs(attr) do
         self[name] =d[name]
    end
end


return layout