local render = {}
local utf8 = require("utf8")

render.mask = {
    index = 0,
    buff = {}
    }


function render.drawToRender()
    local tab = {}
    for id , buff in pairs(render.mask.buff) do
        tab[buff.index] = buff
    end
    
    for _,buff in ipairs(tab) do
        local mask = function()
            for _,func in ipairs(buff.funcs) do
                func()
            end
        end
        love.graphics.stencil(mask,"increment")
        
        for _,member in ipairs(buff.members) do
            --love.graphics.setStencilTest("gequal",member.level)
            member.draw()
            --love.graphics.setStencilTest()
        end
        --[[
        love.graphics.setStencilTest("equal",3)
        love.graphics.rectangle("fill",0,0,2000,2000)
        love.graphics.setStencilTest()]]
    end
    render.mask.buff = {}
    render.mask.index = 0
end

local function dashedLine(ox,oy,tx,ty,pat,type)
    local dist = math.sqrt((tx-ox)^2 + (ty-oy)^2)
    local dx,dy = tx-ox,ty-oy
    local loop = pat[1] + pat[2]
    for i = 0,dist,loop do
        k0 = i/dist
        k1 = (i+pat[1])/dist
        love.graphics.line(ox+k0*dx,oy+k0*dy,ox+k1*dx,oy+k1*dy)
    end
end

local function split(str,keyword)
	local tab={}
	local index=1
	local from=1
	local to=1
	while true do
		if string.sub(str,index,index)==keyword then
			to=index-1
			if from>to then 
				table.insert(tab, "")
			else
				table.insert(tab, string.sub(str,from,to))
			end
			from=index+1
		end
		index=index+1
		if index>string.len(str) then
			if from<=string.len(str) then
				table.insert(tab, string.sub(str,from,string.len(str)))
			end
			return tab
		end
	end
end

local function copy(source,copyto,ifcopyfunction)
	copyto=copyto or {}
	for k, v in pairs(source or {}) do
		if type(v) == "table" then
			copyto[k] = table.copy(v,copyto[k])          
		elseif type(v) == "function" then 
			if ifcopyfunction then
				copyto[k] = v
			end
		else 
			copyto[k] = v
		end
	end
	return copyto
end

local function dividLines(text,font,lineLimit)
    local lines = {}
    local textLen = utf8.len(text)
    local u = 0
    local line = ""
    for i = 1,textLen do
        local char = string.sub(text,utf8.offset(text,i),utf8.offset(text,i+1)-1)
        local cw = font:getWidth(char)
        u = u + cw
        if u> lineLimit then
            table.insert(lines,line)
            u = cw
            line = char
        else
            line = line .. char
        end
            
    end
    table.insert(lines,line)
    return lines
end

local function getAlinePos(aline,w,h,tw,th,line,spacing,lineCount)
    line = line or 1
    spacing = spacing or 0
    lineCount = lineCount or 1
    local x,y
    if aline.x == "left" then
        x = 0
    elseif aline.x == "center" then
        x = w/2 - tw/2
    elseif aline.x == "right" then
        x = w - tw
    end
    if aline.y == "top" then
        y = (line - 1) * (th + spacing) + spacing
    elseif aline.y == "center" then
        y = h/2 - ((th + spacing) * lineCount - spacing )/2 + (th + spacing)* (line - 1)
    elseif aline.y == "bottom" then
        y = h - (th + spacing) * lineCount - spacing + (th + spacing)* (line - 1)
    end
    return x,y
end

local function renderToZone(style,opt,func,pass)
    if opt.noDraw then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = opt.noState and "normal" or opt.state
    local sx = opt.noScale and 1 or style.scale[state].x
    local sy = opt.noScale and 1 or style.scale[state].y
    local offx = opt.noOffset and 0 or style.offset[state].x
    local offy = opt.noOffset and 0 or style.offset[state].y
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx = opt.noRound and 0 or style.roundCorner.rx
    local ry = opt.noRound and 0 or style.roundCorner.ry
    
    
    if not render.mask.buff[opt.container or "screen"] then
        render.mask.index = render.mask.index + 1
        render.mask.buff[opt.container or "screen"] 
            = {funcs = {},members = {},index = render.mask.index ,level = 0}
        if opt.container then
            local buff = render.mask.buff[opt.container or "screen"]
            buff.level = 1
            table.insert(buff.funcs,function()
                love.graphics.rectangle("fill",unpack(opt.container))
            end)
        end
    end
    local buff = render.mask.buff[opt.container or "screen"]
    local renderLevel = buff.level
    if not pass then
        table.insert(buff.funcs,function()
                love.graphics.push()
                love.graphics.translate(xx,yy)
                love.graphics.scale(sx,sy)
                love.graphics.rectangle("fill",0,0,ww,hh,rx,ry)
                love.graphics.pop()
            end)
        renderLevel = renderLevel + 1
    end
    table.insert(buff.members, {
        draw = function()
            love.graphics.push()
            love.graphics.translate(xx,yy)
            love.graphics.scale(sx,sy)
            love.graphics.setScissor(xx-1,yy-1,ww+2,hh+2)
            func(state,ww,hh,rx,ry) 
            love.graphics.setScissor()
            love.graphics.pop()
        end,
        level = renderLevel
        })
end

local function renderToZone(style,opt,func,pass)
    if opt.noDraw then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = opt.noState and "normal" or opt.state
    local sx = opt.noScale and 1 or style.scale[state].x
    local sy = opt.noScale and 1 or style.scale[state].y
    local offx = opt.noOffset and 0 or style.offset[state].x
    local offy = opt.noOffset and 0 or style.offset[state].y
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx = opt.noRound and 0 or style.roundCorner.rx
    local ry = opt.noRound and 0 or style.roundCorner.ry
    
    love.graphics.push()
    love.graphics.translate(xx,yy)
    love.graphics.scale(sx,sy)
    local sx,sy,stx,sty = xx-1,yy-1,xx - 1 + ww+2, yy - 1 + hh+2
    if opt.container then
        sx = sx > opt.container[1] and sx or opt.container[1]
        sy = sy > opt.container[2] and sy or opt.container[2]
        stx = stx < opt.container[1] + opt.container[3] and stx or opt.container[1] + opt.container[3]
        sty = sty < opt.container[2] + opt.container[4] and sty or opt.container[2] + opt.container[4]
    end
    if stx-sx>0 and sty-sy >0 then
        love.graphics.setScissor(sx,sy,stx-sx,sty-sy)
        func(state,ww,hh,rx,ry) 
        love.graphics.setScissor()
    end
    love.graphics.pop()
end


local function renderLine(style,opt,ox,oy,tx,ty)
    local state = opt.state or "normal"
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    love.graphics.setLineWidth(strokeWidth)
    love.graphics.setPointSize(strokeWidth)
    love.graphics.setColor(strokeColor)
    if style.strokeDash.enable then
        local pat = style.strokeDash.pattern
        local type = style.strokeDash.type
        dashedLine(ox,oy,tx,ty,pat,type)
    else
        love.graphics.line(ox,oy,tx,ty)
    end
end

local function renderText(style,opt)
    if opt.noText then return end
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local font = opt.font[state].obj
        local text = opt.text
        local color = style.fontColor[state]
        local cx,cy = w/2,h/2
        local aline = opt.aline or {x = "center",y = "center"}
        local lineSpacing = opt.lineSpacing or 0
        local fw,fh = font:getWidth("W"),font:getHeight("H")   
        local x,y,w,h = rx,ry,w-rx*2,h-ry*2
        
        love.graphics.setColor(color)
        love.graphics.setFont(font)
        
        if string.find(text,"\n") then --自动多行
            local lines = split(text,"\n")
            local lineCount = #lines
            for i = 1, lineCount do
                local line = lines[i]
                local lw = font:getWidth(line)
                local lx,ly = getAlinePos(aline,w,h,lw,fh,i,lineSpacing,lineCount)
                love.graphics.print(line,math.floor(x),math.floor(y))
            end
        elseif opt.multiLine then --自适应多行
            local lines = dividLines(text,font,w)
            local lineCount = #lines
            for i = 1, lineCount do
                local line = lines[i]
                local lw = font:getWidth(line)
                local lx,ly = getAlinePos(aline,w,h,lw,fh,i,lineSpacing,lineCount)
                love.graphics.print(line,math.floor(x),math.floor(y))
            end
        elseif opt.wrapLine then --走马灯
            local lw = font:getWidth(text)
            opt.wrapOffset = opt.wrapOffset or 0
            opt.wrapOffset = opt.wrapOffset +love.timer.getDelta()*(lw+w)/3
            opt.wrapOffset = opt.wrapOffset > lw+w and 0 or opt.wrapOffset 
            love.graphics.print(text,math.floor(x - opt.wrapOffset ),math.floor(y + h/2-fh/2))
            love.graphics.print(text,math.floor(x - opt.wrapOffset + lw+w ),math.floor(y + h/2-fh/2))
        else --单行
            local lw = font:getWidth(text)
            local lx,ly =  getAlinePos(aline,w,h,lw,fh,1,lineSpacing,1)
            love.graphics.print(text,math.floor(lx+x),math.floor(y+ly))
        end    
    end)
end

local function renderRect_fill(style,opt)
    if opt.noFill then return end
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local fillColor = style.fillColor[state]    
        love.graphics.setColor(fillColor)
        love.graphics.rectangle("fill",0,0,w,h,rx,ry)
    end)
end

render.grad_buff = {}

local function renderRect_grad(style,opt,height)
    if opt.noGrad then return end
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local grad = opt.mesh or style.getMesh() 
        local gradColor = style.gradient.color[state]
        local mask = function()
            love.graphics.rectangle("fill",0,0,w,h,rx,ry)
        end
        love.graphics.stencil(mask,"replace",1)
        love.graphics.setStencilTest("equal",1)
        love.graphics.setColor(gradColor)
        love.graphics.draw(grad,0,0,0,w,height or h)
        love.graphics.setStencilTest()  
    end)
end

local function renderRect_stroke(style,opt)
    if opt.noStroke then return end
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local strokeColor = style.strokeColor[state]
        local strokeWidth = style.strokeWidth[state]
        love.graphics.setLineWidth(strokeWidth)
        love.graphics.setPointSize(strokeWidth)
        love.graphics.setColor(strokeColor)
        love.graphics.rectangle("line",0,0,w,h,rx,ry)
    end,true)
end

local function renderRect(style,opt)
    renderRect_fill(style,opt)
    renderRect_grad(style,opt)
    renderRect_stroke(style,opt)
end

local function renderShape(style,opt,shape)
    if opt.toggle == false then return end
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local color = style.strokeColor.active
        love.graphics.setColor(color)
        local s = w>h and h*0.25 or w*0.25
        local cx,cy = w/2,h/2
            if shape == "check" then
                love.graphics.line(cx+s,cy+s,cx-s,cy-s)
                love.graphics.line(cx+s,cy-s,cx-s,cy+s)
            elseif shape == "radio" then
                love.graphics.circle("fill",cx,cy,s)
            elseif shape == "up" then
                love.graphics.polygon("fill",cx,cy-s,cx-s,cy+s,cx+s,cy+s)
            elseif shape == "down" then    
                love.graphics.polygon("fill",cx,cy+s,cx-s,cy-s,cx+s,cy-s)
            elseif shape == "left" then
                love.graphics.polygon("fill",cx-s,cy,cx+s,cy-s,cx+s,cy+s)
            elseif shape == "right" then
                love.graphics.polygon("fill",cx+s,cy,cx-s,cy-s,cx-s,cy+s)
            end
    end)
end

local function renderImage(style,opt)
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local aline = opt.aline or {x = "center",y = "center"}
        local imageColor = style.imageColor[state]    
        local iw,ih = opt.image:getDimensions()
        love.graphics.setColor(imageColor)
        if opt.adjust then
            love.graphics.draw(opt.image,w/2,h/2,opt.rot or 0,w/iw,h/ih,iw/2,ih/2)
        else
            local lx,ly = getAlinePos(aline,w,h,iw,ih)
            love.graphics.draw(opt.image,lx+iw/2,ly+ih/2,opt.rot or 0,1,1,iw/2,ih/2)
        end    
    end)
end


function render.block(style,opt)
    renderRect(style,opt)
end

function render.button(style,opt)
    renderRect(style,opt)
    renderText(style,opt)
end

function render.shape(style,opt)
    renderRect_fill(style,opt)
    renderRect_stroke(style,opt)
    renderShape(style,opt,opt.shape)
end

function render.label(style,opt)
    renderText(style,opt)
end

function render.image(style,opt)
    renderRect_fill(style,opt)
    renderImage(style,opt)
    renderRect_stroke(style,opt)
end

function render.panel(style,opt)
    if opt.withStyle then
        renderRect(style,opt)
    else
        opt.noRound = true
        opt.noScale = true
        opt.noOffset = true
        renderRect_fill(style,opt)
    end
end

function render.progress(style,opt)
    renderRect_fill(style,opt)
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local grad = opt.mesh or style.getMesh()
        local gradColor = style.gradient.color[state]
        love.graphics.setColor(gradColor)
        local gx,gy = 0,0
        local gw,gh = w,h
        if opt.vert then
            gh = gh * opt.value
            if not opt.counter then
                gy = h - gh
            end
        else
            gw = gw*opt.value
            if opt.counter then
                gx = w - gw
            end
        end
        love.graphics.draw(grad,gx,gy,0,gw,gh)    
    end)
    renderRect_stroke(style,opt)
end

function render.slider(style,opt)
    renderRect_fill(style,opt)
    renderRect_stroke(style,opt)
    if not opt.noValue then
        renderText(style,opt)
    end
end


function render.text(style,opt)
    renderRect(style,opt)
    renderToZone(style,opt,function(state,w,h,rx,ry)
        local g = opt.grid
        if opt.debug then
            love.graphics.setColor(0.2,0.2,1,1)
            for i = 1,g.count do
                local c = g.cells[i]
                love.graphics.rectangle("line",g.x + c.x,g.y,c.w,g.h)
            end
            love.graphics.setColor(1,0,0,0.3)
            love.graphics.rectangle("fill",g.x,g.y,g.w,g.h)
        end
        local color = style.fontColor.normal
        love.graphics.setColor(color)
        local font = opt.font.normal.obj
        love.graphics.setFont(font)
        for i = 1,g.count do
            local c = g.cells[i]
            love.graphics.print(c.char, math.floor(g.x + c.x),math.floor(g.y))
        end
        
        local c = opt.cursor
        if opt.editing then
            c.timer = c.timer - love.timer.getDelta()
            if c.timer<0 then
                c.timer = c.cd
            end
            love.graphics.setColor(1,1,1,c.timer/c.cd)
            if opt.cursor.mode == "insert" then
                love.graphics.line(g.x + c.x,g.y,g.x+c.x,g.y+g.h)
            else
                love.graphics.rectangle("fill",g.x+c.x,g.y,c.w,g.h)
            end
            
        end    
    end)
end

function render.tip(style,opt)
    renderRect(style,opt)
    renderText(style,opt)
end

function render.frame(style,opt)
    renderRect_fill(style,opt)
    if not opt.folded then
        renderToZone(style,opt,function(state,w,h,rx,ry)
            renderLine(style,opt,0,opt.title_h,w,opt.title_h)  
        end,true)
    end
    renderRect_grad(style,opt,opt.title_h)
    renderRect_stroke(style,opt)
end

function render.canvas(style,opt)
    renderToZone(style,opt,function(state,w,h,rx,ry)
        love.graphics.push()
        love.graphics.translate(-opt.offsetX,-opt.offsetY)
        opt.render()   
        love.graphics.pop()
    end)
end

function render.tab(style,opt)
    --renderRect(style,opt)
end

return render