local style = {}

style.noStencil = true

style.strokeColor = {
    normal = {1,1,1,1},
    active = {1,1,0.5,1},
    hovered = {0.5,0.7,1,1},
    disabled = {0.5,0.5,0.5,1}
    }

style.strokeWidth = {
    normal = 1,
    active = 1,
    hovered = 2,
    disabled = 1
}

style.fillColor = {
    normal = {0.3,0.5,1,1},
    active = {1,0.3,0.5,1},
    hovered = {0.5,0.7,1,1},
    disabled = {0.5,0.5,0.5,1}
    }

style.gradient = {
    enable = true,
    type = "horizonal", -- vertical horizonal  radius
    color = {
        normal = {1,1,1,1},
        active = {1,1,1,1},
        hovered = {1,1,1,1},
        disabled = {0.5,0.5,0.5,0}
        },
    color0 = {0.5,0.5,0.5,1},
    color1 = {0.5,0.5,1,1},
}

style.scale = {
    normal = {x = 1,y = 1},
    active = {x = 1,y = 1},
    hovered = {x = 1,y = 1},
    disabled = {x = 1,y = 1},
}

style.offset = {
    normal = {x = 0,y = 0},
    active = {x = 2,y = 2},
    hovered = {x = 0,y = 0},
    disabled = {x = 0,y = 0},
    }

style.strokeDash = {
    enable = false,
    pattern = {5,2},
    type = "line", -- dot
    }

style.roundCorner = {rx = 10,ry = 10}

style.font = {
    normal = {path = "",size = 14},
    active = {path = "",size = 13},
    hovered = {path = "",size = 15},
    disabled = {path = "",size = 14},
    }

style.fontColor = {
    normal = {0.9,0.9,0.9,1},
    active = {1,1,1,1},
    hovered = {1,1,1,1},
    disabled = {0.3,0.3,0.3,1}
}

style.imageColor = {
    normal = {0.9,0.9,0.9,1},
    active = {1,1,0.5,1},
    hovered = {1,1,1,1},
    disabled = {0.3,0.3,0.3,1}
}


style.buff = {}
local attr = {"strokeColor","strokeWidth","fillColor",
    "gradient","scale","offset","strokeDash","roundCorner",
    "fontPath","fontColor","fontScale"}
function style.push()
    local data = {}
    for i,name in ipairs(attr) do
        data[name] = type(style[name]) == "table" and table.copy(style[name]) or style[name]
    end
    table.insert(style.buff,data)
end

function style.pop()
    for k,v in pairs(style.buff[#style.buff]) do
        style[k] = v
    end
    style.buff[#style.buff] = nil
end

function style.newFont()
    local state = {"normal","active","hovered","disabled"}
    for i,s in ipairs(state) do
        local t = style.font[s]
        if t.path and t.path ~= "" then
            t.obj = t.obj or love.graphics.newFont(t.path,t.size)
        else
            t.obj = t.obj or love.graphics.newFont(t.size)
        end
        
    end
    
    return style.font
end

local function newMesh(type,color0,color1)
    local mesh
    local r0,g0,b0,a0 = unpack(color0)
    local r1,g1,b1,a1 = unpack(color1)
    if type == "radius" then
        mesh = love.graphics.newMesh({
            {0.5,0.5,0.5,0.5,r0,g0,b0,a0},
            {0,0,0,0,r1,g1,b1,a1},
            {1,0,1,0,r1,g1,b1,a1},
            {1,1,1,1,r1,g1,b1,a1},
            {0,1,0,1,r1,g1,b1,a1},
            {0,0,0,0,r1,g1,b1,a1},
                },"fan")
    elseif type == "vertical" then
        mesh = love.graphics.newMesh({
            {0,0,0,0,r0,g0,b0,a0},
            {1,0,1,0,r1,g1,b1,a1},
            {0,1,0,1,r0,g0,b0,a0},
            {1,1,1,1,r1,g1,b1,a1},
                },"strip")
    elseif type == "horizonal" then
        mesh = love.graphics.newMesh({
            {0,0,0,0,r0,g0,b0,a0},
            {1,0,1,0,r0,g0,b0,a0},
            {0,1,0,1,r1,g1,b1,a1},
            {1,1,1,1,r1,g1,b1,a1},
                },"strip")
    end
    return mesh
end

function style.getMesh()
    if style.gradient.mesh then
        return style.gradient.mesh
    else
        local t = style.gradient
        t.mesh = newMesh(t.type,t.color0,t.color1)
        return t.mesh
    end
end

return style