local utf8 = require 'utf8'

local function updateGrid(opt)
    local style = opt.style
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = opt.noState and "normal" or opt.state
    local sx = opt.noScale and 1 or style.scale[state].x
    local sy = opt.noScale and 1 or style.scale[state].y
    local offx = opt.noOffset and 0 or style.offset[state].x
    local offy = opt.noOffset and 0 or style.offset[state].y
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx = opt.noRound and 0 or style.roundCorner.rx
    local ry = opt.noRound and 0 or style.roundCorner.ry
    local font = opt.font.normal.obj
    local tw = font:getWidth(opt.text)
    local th = font:getHeight()
    opt.grid = {
        ox = xx,
        oy = yy,
        x =  rx,
        y =  h/2 - th/2,
        w = ww-rx*2,
        h = th,
        count = utf8.len(opt.text)
    }

    opt.aline = opt.aline or "right"
    opt.grid.cells = {}
    local grid = opt.grid
    local cells = opt.grid.cells
    local count = grid.count
    local text = opt.text
    local a = opt.aline
    a = opt.editing and "center" or a
    local offset
    if a == "left" then
        offset = 0
    elseif a == "right" then
        offset = grid.w - tw
    elseif a == "center" then
        offset = grid.w/2 - tw/2
    elseif a == "free" then
        offset = opt.offset_text or 0
    end
    
    for i = 1,count do
        cells[i] = {}
        local cell = cells[i]
        local l,r = utf8.offset(text,i),utf8.offset(text,i+1) -1
        local char = opt.text:sub(l,r)
        local cw = font:getWidth(char)
        cell.l = l
        cell.r = r
        cell.char = char
        cell.w = cw
        cell.x = offset
        offset = offset + cw
    end
    
end

local function newCursor(opt)
    opt.cursor = {
        pos = 1,
        mode = "insert",
        timer = 0,
        cd = 1
        }
end

local function addText(opt,text)
    local cells = opt.grid.cells
    local pos = opt.cursor.pos
    local left = cells[pos-1]
    local right = cells[pos]
    local lt = left and string.sub(opt.text,1,left.r) or ""
    local rt = right and string.sub(opt.text,right.l,-1) or ""
    opt.text = lt .. text .. rt
    opt.cursor.pos = pos + 1
    updateGrid(opt)
end

local function delText(opt,pos)
    local cells = opt.grid.cells
    local pos = opt.cursor.pos
    local left = cells[pos-2]
    local right = cells[pos]
    if opt.cursor.mode == "replace" then
        left = cells[pos-1]
        right = cells[pos + 1]
    end
    local lt = left and string.sub(opt.text,1,left.r) or ""
    local rt = right and string.sub(opt.text,right.l,-1) or ""
    opt.text = lt .. rt
    opt.text = opt.text == "" and " " or opt.text
    opt.cursor.pos = pos - 1
    updateGrid(opt)
end

local function replaceText(opt,text)
    local cells = opt.grid.cells
    local pos = opt.cursor.pos
    local left = cells[pos-1]
    local right = cells[pos+1]
    local lt = left and string.sub(opt.text,1,left.r) or ""
    local rt = right and string.sub(opt.text,right.l,-1) or ""
    opt.text = lt ..text.. rt
    opt.cursor.pos = pos + 1
    updateGrid(opt)
end

local function updateCursorPos(opt)
    local c = opt.cursor
    local g = opt.grid
    local cells = g.cells
    c.pos = c.pos < 1 and 1 or c.pos
    c.pos = c.pos > g.count + 1 and g.count + 1 or c.pos
    local p = c.pos
    local cell = cells[p]
    if cell then
        c.x = cell.x
        c.w = cell.w
    else
        c.x = cells[g.count].x + cells[g.count].w 
        c.w = cells[g.count].w
    end
end

local function updateCursor(opt,mx,my)
    local c = opt.cursor
    local g = opt.grid
    local cells = g.cells
    if mx< g.x + cells[1].x then
        c.pos = 1
    elseif mx> g.x+ cells[g.count].x + cells[g.count].w then
        c.pos = g.count + 1
    else
        for i , cell in ipairs(cells) do
            if mx<g.x + cell.x + cell.w then
                c.pos = i
                break
            end
        end
    end
    updateCursorPos(opt)
end




local function keyConf(opt,key)
    if key == "backspace" then
        delText(opt)
    elseif key == "right" then
        opt.cursor.pos = opt.cursor.pos + 1
        updateCursorPos(opt)
    elseif key == "left" then
        opt.cursor.pos = opt.cursor.pos - 1
        updateCursorPos(opt)
    elseif key == "insert" then
        opt.cursor.mode = opt.cursor.mode == "insert" and "replace" or "insert"
    elseif key == "return" then
        opt.editing = false
    end
end

local name = "text"
return function(core,opt)
    opt.id = opt.id or tostring(name.."_"..tostring(opt))
    opt.font = opt.font or core.style.newFont()
    opt.state = opt.state or "normal"
    ---interact
    opt.noScale = true
    opt.submit = nil
    opt.editing = false
    if opt.disabled then
        opt.state = "disabled"
    else
        core:updateState(opt)
    end
    opt.editing = core:grabFocus(opt) 
    opt.active = opt.editing
    if opt.editing and core.text then
        if opt.cursor.mode == "insert" then
            addText(opt,core.text)
        else
            replaceText(opt,core.text)
        end
        updateCursorPos(opt)
    end
    
    if opt.editing and core.key then
        if core.key == "return" then
            core:releaseFocus()
            opt.submit = opt.text
        else
            keyConf(opt,core.key)
        end
        updateCursorPos(opt)
    end
    
    
    ---draw
    core:applyLayout(opt)
    opt.draw = opt.draw or core.render[name]
    opt.style = opt.style or core.style
    
    if not opt.buff or opt.dirty then
        updateGrid(opt)
    end
    if not opt.cursor then
        newCursor(opt)
    end
    if core:isActive(opt) then
        updateCursor(opt,core.mouse_x - opt.grid.ox,core.mouse_y - opt.grid.oy)
    end


    if core:gotFocus(opt) then
        updateGrid(opt)
    end
    
    if core:hadFocus(opt) then
        updateGrid(opt)
    end
    
    core:registerDraw(opt.draw ,{opt.style,opt})
    core:updateBound(opt)
    return opt
end