local name = "frame"

local function update(core,opt)
    local sx,sy = 1,1
    local offx,offy = 0,0
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local style = opt.style or core.style
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    opt.noOffset = true
    opt.noScale = true
    
    if opt.resizable then
        if not core.mouse_down then
            opt.grab = nil
        elseif opt.resize_btn and opt.resize_btn.state == "active" then
            opt.grab = "resize" 
            if opt.layout then opt.layout.disable = true end
        end
    end
    
    if opt.movable then
        if not core.mouse_down then
            opt.grab = nil
        elseif opt.state == "active" then
            opt.grab = "move"
            if opt.layout then opt.layout = nil end
        end
    end
    
    if opt.grab == "resize" then
        opt.w = opt.w + core.mouse_dx
        opt.h = opt.h + core.mouse_dy
        opt.w = math.max(opt.w,opt.title_label.h*4)
        opt.h = math.max(opt.h,opt.title_label.h+ry)
        if opt.parent then
            opt.parent.w = opt.w
            opt.parent.h = opt.h
            opt.parent.layout = nil
            opt.w = math.max(opt.w,opt.parent.tab_w)
        end
    elseif opt.grab == "move" then
        opt.x = opt.x + core.mouse_dx
        opt.y = opt.y + core.mouse_dy
    end
    x,y,w,h = opt.x,opt.y,opt.w,opt.h
    xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    
    local font = opt.font.normal.obj
    local text = opt.text
    local fw,fh = font:getWidth("W"),font:getHeight("H")
    local tw = font:getWidth(opt.title)
    local ss = fh+2
    opt.title_h = ss
    if not opt.title_label then
        opt.title_label = {
            x = xx + rx + 1,
            y = yy + 1,
            w = ww,
            h = fh,
            text = opt.title,
            aline = {x = "left",y = "center"},
            font = {
                normal= opt.font.normal,
                hovered= opt.font.normal,
                active= opt.font.normal,
                disabled = opt.font.normal
                }
        }
        opt.close_btn = {
            x = xx + ww  - ss,
            y = yy + 1,
            w = fh,
            h = fh,
            shape = "check"
        }
        
        opt.min_btn = {
            x = xx + ww  - ss*2 - 1,
            y = yy + 1,
            w = fh,
            h = fh,
            shape =  "up"
        }
        
        opt.resize_btn = {
            x = xx + ww - 5,
            y = yy + hh - 5,
            w = 5,
            h = 5,
            shape = "null",
            noDraw = true,
            noOffset = true
        }
        
        if opt.userDraw then
            opt.canvas = {
                x = xx,
                y = yy + fh + 2,
                w = ww,
                h = hh - fh - 2,
                render = opt.userDraw
                }
            
        end
    elseif not opt.noHead then
        local tl = opt.title_label
        local mb = opt.min_btn
        local cb = opt.close_btn
        tl.x = xx + rx + 1
        tl.y = yy + 1
        core:doWidget("label",tl)
        local x = xx + ww
        if opt.close then
            cb.x = x - ss
            cb.y = yy + 1
            core:doWidget("shape",cb)
            x = x - ss - 1
        end
        if opt.min then
            mb.x = x - ss
            mb.y = yy + 1
            core:doWidget("shape",mb)
        end
        
        if opt.min_btn.hit then
            local btn = opt.min_btn
            if btn.shape == "up" then
                btn.shape = "down"
                btn.frame_h = hh
                opt.h = opt.title_h
                opt.folded = true
            elseif btn.shape == "down" then
                btn.shape = "up"
                opt.h = btn.frame_h
                btn.frame_h = nil
                opt.folded = false
            end
        end
        if opt.close_btn.hit then
            opt.hide = true
        end
       
    end
    
    if opt.resizable then
        local rb = opt.resize_btn
        rb.x = xx + ww - 5
        rb.y = yy + hh - 5
        core:doWidget("shape",rb)
    end
    if opt.userDraw then
        opt.canvas.x = xx
        opt.canvas.y = yy + fh + 2
        core:doWidget("canvas",opt.canvas)
    end
    
    if opt.window then
        local sw = 10
        opt.window.x = opt.window.x or 0
        opt.window.y = opt.window.y or 0
        if opt.window.w > ww  and not opt.noSlider then
            if opt.slider_x then
                opt.slider_x.x = xx
                opt.slider_x.y = yy + hh - ss
                opt.slider_x.w = ww - sw
                opt.slider_x.max = opt.window.w - opt.w
            else
                opt.slider_x  = {
                    x = xx,
                    y = yy+hh-ss,
                    w = ww - sw,
                    h = ss,
                    min = 0,
                    max = opt.window.w - opt.w,
                    value = opt.window.x,
                    noValue = true
                }
            end
            core:doWidget("slider",opt.slider_x)
            opt.window.x = opt.slider_x.value
        end
        
        local wh = hh - fh - 2
        if opt.window.h > wh and not opt.noSlider then
            if opt.slider_y then
                opt.slider_y.x = xx + ww - ss
                opt.slider_y.y = yy + fh + 2
                opt.slider_y.h = wh - sw
                opt.slider_y.max = opt.window.h - wh
            else
                opt.slider_y  = {
                    x = xx + ww - ss,
                    y = yy+fh+ 2,
                    w = ss,
                    h = wh -sw,
                    min = 0,
                    max = opt.window.h - wh,
                    value = opt.window.y,
                    noValue = true,
                    vert = true
                }
            end
            core:doWidget("slider",opt.slider_y)
            opt.window.y = opt.slider_y.value
            
        end
    end
    
    
    if not opt.folded then
        local wx,wy,ww,wh = xx+1,yy+fh+1,ww-2,hh-2-fh
        local container = {xx,yy,ww,hh}
        opt.title_label.container = container
        if opt.canvas then
            opt.canvas.container = container
            opt.canvas.w = ww
            opt.canvas.h = wh
            if opt.slider_x then
                opt.canvas.offsetX = opt.slider_x.value
            end
            if opt.slider_y then
                opt.canvas.offsetY = opt.slider_y.value
            end
        end
        
        
        if opt.window then
            wx = wx - opt.window.x
            wy = wy - opt.window.y
            ww = opt.window.w
            wh = opt.window.h
        end
        
        if opt.cell then
            local layout = core.layout
            layout:push()
            layout:setWindow(wx,wy,ww,wh)
            layout:setCell(unpack(opt.cell))
            for i,child in ipairs(opt.children or {}) do
                child.container = container
                core:doWidget(child.type,child)
            end
            layout:pop()
        elseif opt.grid then
            local layout = core.layout
            layout:push()
            layout:setWindow(wx,wy,ww,wh)
            layout:cutGrid(unpack(opt.grid))
            for i,child in ipairs(opt.children or {}) do
                child.container = container
                core:doWidget(child.type,child)
            end
            layout:pop()
        else
            for i,child in ipairs(opt.children or {}) do
                child.container = container
                core:doWidget(child.type,child)
            end
        end
    end
end




return function(core,opt)
    opt.id = opt.id or tostring(name.."_"..tostring(opt))
    core:applyLayout(opt)
    ---interact
    opt.font = opt.font or core.style.newFont()
    opt.state = opt.state or "normal"
    opt.noScale = true
    opt.noOffset = true
    if opt.disabled or opt.hide then
        opt.state = "disabled"
    else
        if not opt.noHead then
            core:updateState(opt)
        end
        update(core,opt)
    end
    
    ---children
    
    
    ---draw
    
    opt.draw = opt.draw or core.render[name]
    opt.style = opt.style or core.style
    if not opt.hide then
        core:registerDraw(opt.draw ,{opt.style,opt})
    end
    return opt
end